@@ -2,6 +2,7 @@
 
 {-# LANGUAGE BangPatterns        #-}
 {-# LANGUAGE LambdaCase          #-}
+{-# LANGUAGE MultiWayIf          #-}
 {-# LANGUAGE NamedFieldPuns      #-}
 {-# LANGUAGE RecordWildCards     #-}
 {-# LANGUAGE ScopedTypeVariables #-}
@@ -41,6 +42,18 @@ module GHC.TcPlugin.API.TyConSubst (
   , splitTyConApp_upTo
   ) where
 
+-- Word64Map is available:
+--
+--  - from 9.10.1 onwards
+--  - in 9.8.4
+--  - in 9.6.7
+--
+-- This is unusual: whether a module is exposed depends on the MINOR version
+#define HAS_WORD64MAP \
+    (MIN_VERSION_ghc(9,9,0) || \
+    (MIN_VERSION_ghc(9,8,4) && !(MIN_VERSION_ghc(9,9,0))) || \
+    (MIN_VERSION_ghc(9,6,7) && !(MIN_VERSION_ghc(9,7,0))))
+
 -- base
 import Data.Bifunctor
   ( Bifunctor(first, second) )
@@ -63,7 +76,7 @@ import Data.Sequence
 import qualified Data.Sequence as Seq
 
 -- ghc
-#if MIN_VERSION_ghc(9,5,0)
+#if HAS_WORD64MAP
 import qualified GHC.Data.Word64Map.Strict as Word64Map
 #else
 import qualified Data.IntMap.Strict as IntMap
@@ -111,7 +124,7 @@ data TyConSubst = TyConSubst {
 --   derive any additional equalities from them (i.e. if, say, we know that
 --   @x ~ T1@ and @x ~ T2@, we will not attempt to use the fact that this means
 --   that @T1 ~ T2@, nor any derived conclusions thereof). We /will/ however
---   try to apply the canononical equalities as often as is necessary (e.g.,
+--   try to apply the canonical equalities as often as is necessary (e.g.,
 --   first applying @x ~ T y@, then applying @y ~ T2@).
 --
 -- We solve this problem by constructing a 'TyConSubst': a possibly
@@ -289,7 +302,7 @@ classify = go mempty
 -- o We keep doing this until we can make no more progress.
 --
 -- The functions for working with 'TyConSubst' take the variable equivalence
--- classes into acocunt, so we do not need to do that here.
+-- classes into account, so we do not need to do that here.
 --
 -- Two observations:
 --
@@ -391,7 +404,6 @@ isCanonicalVarEq = \case
       Just (cc_tyvar, cc_rhs, cc_eq_rel)
     CFunEqCan { cc_fsk, cc_fun, cc_tyargs } ->
       Just (cc_fsk, mkTyConApp cc_fun cc_tyargs, NomEq)
-    _otherwise    -> Nothing
 #elif __GLASGOW_HASKELL__ < 907
     CEqCan { cc_lhs, cc_rhs, cc_eq_rel }
       | TyVarLHS var <- cc_lhs
@@ -399,8 +411,6 @@ isCanonicalVarEq = \case
       | TyFamLHS tyCon args <- cc_lhs
       , Just var            <- getTyVar_maybe cc_rhs
       -> Just (var, mkTyConApp tyCon args, NomEq)
-    _otherwise
-      -> Nothing
 #else
     CEqCan eqCt
       | TyVarLHS var <- lhs
@@ -412,9 +422,18 @@ isCanonicalVarEq = \case
         lhs = eq_lhs eqCt
         rhs = eq_rhs eqCt
         rel = eq_eq_rel eqCt
-    _otherwise
-      -> Nothing
 #endif
+    -- Deal with CNonCanonical, which are produced by ghc-tcplugin-api
+    -- in GHC 9.0 and below due to 'unflattenCts'.
+    ct@(CNonCanonical {})
+      | EqPred rel lhs rhs <- classifyPredType (ctPred ct)
+      -> if | Just tv <- getTyVar_maybe lhs
+            -> Just (tv, rhs, rel)
+            | Just tv <- getTyVar_maybe rhs
+            -> Just (tv, lhs, rel)
+            | otherwise
+            -> Nothing
+    _ -> Nothing
 
 {-------------------------------------------------------------------------------
   Internal auxiliary
@@ -498,7 +517,7 @@ minViewUniqSet s =
           ( unsafeCoerce :: IntMap.IntMap a -> UniqFM a a )
 #endif
         ) <$>
-#if MIN_VERSION_ghc(9,5,0)
+#if HAS_WORD64MAP
       Word64Map.minView
 #else
       IntMap.minView
